/*
 * Copyright (c) 1995 Stephen J. Williams
 */
#ident "$Id: bppmanip.cc,v 1.1 1995/10/27 03:22:59 steve Exp $"
/*
 * $Log: bppmanip.cc,v $
 * Revision 1.1  1995/10/27 03:22:59  steve
 *  Add bpp to CVS.
 *
 *
 */


# include  <iostream.h>
# include  <string.h>
# include  <fcntl.h>
# include  <stdlib.h>
# include  <errno.h>
# include  "bpp.h"
# include  <sys/ioctl.h>
# include  <assert.h>

const char notice[] = "Copyright (c) 1995 Picture Elements, Inc.\n";
const char white[] = " \t\f\n";

static bool done = false;
static int  device = -1;

typedef void (*cmd_fun)(unsigned argc, char*argv[]);
cmd_fun lookup(const char*name);

main(int argc, char*argv[])
{
      if (argc != 2) {
	    cerr << "Usage: " << argv[0] << " <device>" << endl;
	    return -1;
      }

      cout << "\n\n" << notice << "\n";

      device = open(argv[1], O_RDWR, 0);
      if (device == -1) {
	    cerr << "Error (" << strerror(errno) << ") opening device `" <<
	      argv[1] << ".'" << endl << endl;
	    return 1;
      }

      while (! done) {

	      // Get a line from the terminal
	    char buffer[1024];
	    cout << "* " << flush;
	    cin.getline(buffer, sizeof buffer);

	      // Chop off comments.
	    char*cp = buffer;
	    cp += strcspn(cp, "#");
	    *cp = 0;

	      // Chop the line into words
	    char*argv[256];
	    unsigned argc = 0;
	    cp = buffer;
	    while (*cp) {

		  cp += strspn(cp, white);
		  if (*cp == 0) break;

		  argv[argc++] = cp;
		  cp += strcspn(cp, white);

		  if (*cp) *(cp++) = 0;
	    }

	    if (argc > 0) lookup(argv[0])(argc, argv);
      }

      return 0;
}


static void no_such_command(unsigned, char*argv[])
{
      cout << argv[0] << ": No such command. Try `help' for help.\n";
}

static void help(unsigned argc, char*argv[])
{
      cout << notice <<
	"This is a program to manipulate the BPP driver.\n"
	"The `#' character starts a comment, and may appear anywhere\n"
	"in a line. Any text that is not white space is chopped into\n"
	"words, and the first word used to identify the command to execute.\n"
	"\n"
	"Supported commands are:\n"
	"\n"
	"    help\n"
	"    ?\n"
	"    quit\n"
	"    read           -- read bytes from the port.\n"
	"    show           -- Show the status of all the port pins.\n"
	"    set <pins>...  -- Set the specified pins to high or low.\n"
	"    change <pins>... -- change the state of the named pins.\n"
	"    data [byte]    -- Get (or set) the data pins.\n"
	"    input          -- Set the data pins to input mode.\n"
	"    output         -- Turn pins to output.\n"
	"    ECP            -- Negotiate ECP mode. Show all the events.\n"
	"    ecpsend        -- Send bytes the ECP way.\n"
	"    nibble         -- Negotiate nibble mode.\n"
	"    terminate      -- Terminate from a IEEE 1284 mode.\n"
	;
}

static void quit(unsigned, char*[])
{
      done = true;
}

static void initpins(unsigned, char*[])
{
      const long pins = BPP_PP_nInit|BPP_PP_nStrobe|BPP_PP_nAutoFd;
      int rc = ioctl(device, BPP_PUT_PINS, pins);
      if (rc == -1)
	cerr << "Error `" << strerror(errno) << "' setting pins." << endl;
}

static void showpins(unsigned, char*[])
{
      int rc = ioctl(device, BPP_GET_PINS);
      if (rc == -1) {
	    cerr << "Error `" << strerror(errno) << "' getting pins." <<
	      endl;
	    return;
      }

      cout << "==== HOST pin values:" << endl;
      cout << " nStrobe: " << ((rc&BPP_PP_nStrobe) ? "HIGH" : "LOW") << endl;
      cout << " nAutoFd: " << ((rc&BPP_PP_nAutoFd) ? "HIGH" : "LOW") << endl;
      cout << " nInit:   " << ((rc&BPP_PP_nInit) ? "HIGH" : "LOW") << endl;
      cout << " nSelectIn: " << ((rc&BPP_PP_nSelectIn) ? "HIGH" : "LOW") << endl;

      cout << "==== PERIPHERAL pin values:" << endl;
      cout << " nAck: " << ((rc&BPP_GP_nAck) ? "HIGH" : "LOW") << endl;
      cout << " Busy: " << ((rc&BPP_GP_Busy) ? "HIGH" : "LOW") << endl;
      cout << " PError: " << ((rc&BPP_GP_PError) ? "HIGH" : "LOW") << endl;
      cout << " Select: " << ((rc&BPP_GP_Select) ? "HIGH" : "LOW") << endl;
      cout << " nFault: " << ((rc&BPP_GP_nFault) ? "HIGH" : "LOW") << endl;
}

static void changepins(unsigned argc, char*argv[])
{
      if (argc == 1) {
	    cerr << "Usage: " << argv[0] << " <pin name>..." << endl;
	    cerr << "Where pin names are: "
	            "  nStrobe, nAutoFd, nInit, nSelectIn" << endl;
	    return;
      }

      int pins = ioctl(device, BPP_GET_PINS);
      if (pins == -1) {
	    cerr << "Error `" << strerror(errno) << "' getting pins." <<
	      endl;
	    return;
      }

      pins &= 0x000f;

      for (unsigned idx = 1 ;  idx < argc ;  idx += 1) {
	    if (strcmp(argv[idx], "nStrobe") == 0)
	      pins ^= BPP_PP_nStrobe;
	    else if (strcmp(argv[idx], "nAutoFd") == 0)
	      pins ^= BPP_PP_nAutoFd;
	    else if (strcmp(argv[idx], "nInit") == 0)
	      pins ^= BPP_PP_nInit;
	    else if (strcmp(argv[idx], "nSelectIn") == 0)
	      pins ^= BPP_PP_nSelectIn;
	    else
	      cerr << "Invaline pin name: " << argv[idx] << endl;
      }

      int rc = ioctl(device, BPP_PUT_PINS, pins);
      if (rc == -1)
	cerr << "Error `" << strerror(errno) << "' setting pins." << endl;
}


void setpins(unsigned argc, char*argv[])
{
      if (argc == 1) {
	    cerr << "Usage: " << argv[0] << " <pin name>..." << endl;
	    cerr << "Where pin names are:  nStrobe, nAutoFd, nInit, nSelectIn\n";
	    cerr << "Prepend a ! to the name to set the pin LOW." << endl;
	    return;
      }

      int pins = ioctl(device, BPP_GET_PINS);
      if (pins == -1) {
	    cerr << "Error `" << strerror(errno) << "' getting pins." <<
	      endl;
	    return;
      }

      pins &= 0x000f;

      for (unsigned idx = 1 ;  idx < argc ;  idx += 1) {
	    if (strcmp(argv[idx], "nStrobe") == 0)
	      pins |= BPP_PP_nStrobe;
	    else if (strcmp(argv[idx], "nAutoFd") == 0)
	      pins |= BPP_PP_nAutoFd;
	    else if (strcmp(argv[idx], "nInit") == 0)
	      pins |= BPP_PP_nInit;
	    else if (strcmp(argv[idx], "nSelectIn") == 0)
	      pins |= BPP_PP_nSelectIn;
	    else if (strcmp(argv[idx], "!nStrobe") == 0)
	      pins &= ~BPP_PP_nStrobe;
	    else if (strcmp(argv[idx], "!nAutoFd") == 0)
	      pins &= ~BPP_PP_nAutoFd;
	    else if (strcmp(argv[idx], "!nInit") == 0)
	      pins &= ~BPP_PP_nInit;
	    else if (strcmp(argv[idx], "!nSelectIn") == 0)
	      pins &= ~BPP_PP_nSelectIn;
	    else
	      cerr << "Invaline pin name: " << argv[idx] << endl;
      }

      int rc = ioctl(device, BPP_PUT_PINS, pins);
      if (rc == -1)
	cerr << "Error `" << strerror(errno) << "' setting pins." << endl;
}

void datapins(unsigned argc, char*argv[])
{
      if (argc > 2) {
	    cerr << "Usage: " << argv[0] << " [byte]" << endl;
	    return;
      }

      if (argc == 1) {
	    int byte = ioctl(device, BPP_GET_DATA);
	    if (byte == -1)
	      cerr << "Error `" << strerror(errno) <<
		"' reading data pins." << endl;
	    else
	      cout << "DATA (hex): " << hex << (byte & 0xff) << dec << endl;

      } else {
	    const int byte = strtoul(argv[1], 0, 0);
	    int rc = ioctl(device, BPP_PUT_DATA, byte & 0xff);
	    if (rc == -1)
	      cerr << "Error `" << strerror(errno) <<
		"' setting data pins." << endl;
      }
}

void inout(unsigned argc, char*argv[])
{
      int rc;

      if (strcmp(argv[0], "input") == 0)
	rc = ioctl(device, BPP_SET_INPUT, 1);
      else
	rc = ioctl(device, BPP_SET_INPUT, 0);


      if (rc == -1)
	cerr << "Error `" << strerror(errno) <<
	  "' setting data pins." << endl;
}


const int PP_MASK=BPP_PP_nStrobe|BPP_PP_nAutoFd|BPP_PP_nInit|BPP_PP_nSelectIn;
const int GP_MASK = 0x1f00;

static void ECP(unsigned, char*[])
{
      int rc;
      int pins = ioctl(device, BPP_GET_PINS);

      cout << "--> Negotiate ECP mode." << endl;

      cout << "Event 0: ";
      if ( (pins & PP_MASK) != (BPP_PP_nInit|BPP_PP_nStrobe|BPP_PP_nAutoFd) ) {
	    cout << "Not in compatibility mode!" << endl;
	    return;
      }
      cout << "OK" << endl;

        // Perform event 1
      rc = ioctl(device, BPP_PUT_DATA, 0x10);
      assert( rc != -1 );
      pins = (pins | BPP_PP_nSelectIn) & ~BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      assert( rc != -1 );
      cout << "Event 1: DONE" << endl;

        // Watch for Event 2 response from the peripheral
      cout << "Event 2: " << flush;
      pins = ioctl(device, BPP_GET_PINS);
      if ( ! (pins & BPP_GP_PError) ) {
	    cout << "PError LOW!" << endl;
	    return;
      }

      if (pins & BPP_GP_nAck) {
	    cout << "nAck HIGH!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_nFault) ) {
	    cout << "nFault LOW!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_Select) ) {
	    cout << "Select LOW!" << endl;
	    return;
      }

      cout << "OK" << endl;

        // Perform Event 3
      pins &= ~BPP_PP_nStrobe;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      cout << "Event 3: DONE" << endl;

        // Perform Event 4
      pins |= BPP_PP_nStrobe|BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      cout << "Event 4: DONE" << endl;

	// Event 5 is transient.
      cout << "(Event 5 is transient)" << endl;

        // Check for Event 6
      cout << "Event 6: " << flush;
      pins = ioctl(device, BPP_GET_PINS);

      if (pins & BPP_GP_PError) {
	    cout << "PError HIGH!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_nAck) ) {
	    cout << "nAck LOW!" << endl;
	    return;
      }

      if (pins & BPP_GP_Busy) {
	    cout << "Busy HIGH" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_nFault) ) {
	    cout << "nFault LOW!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_Select) ) {
	    cout << "Select LOW!" << endl;
	    return;
      }

      cout << "OK" << endl;

      cout << "--> Setup Phase" << endl;

        // Perform Event 30
      pins &= ~BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      cout << "Event 30: DONE" << endl;

        // Check for Event 31
      cout << "Event 31: " << flush;

      pins = ioctl(device, BPP_GET_PINS);
      if ( ! (pins & BPP_GP_PError) ) {
	    cout << "PError LOW!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_nAck) ) {
	    cout << "nAck LOW!" << endl;
	    return;
      }

      if (pins & BPP_GP_Busy) {
	    cout << "Busy HIGH!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_Select) ) {
	    cout << "Select LOW!" << endl;
	    return;
      }

      if (pins & BPP_GP_nFault)
	cout << "(nFault HIGH) OK" << endl;
      else
	cout << "(nFault LOW) OK" << endl;

      cout << "--> ECP Forward Phase" << endl;
}

static void ecpsend(unsigned argc, char*argv[])
{
      int rc;
      int pins = ioctl(device, BPP_GET_PINS);
      assert( pins != -1 );

      if (! (pins & BPP_PP_nSelectIn)) {
	    cout << "1284 not active!" << endl;
	    return;
      }

      for (unsigned idx = 1 ;  idx < argc ;  idx += 1) {

	    cout << "--> Sending an ECP byte" << endl;

	    cout << "Event 33: " << flush;
	    pins = ioctl(device, BPP_GET_PINS);
	    assert( pins != -1 );

	    if ( ! (pins & BPP_GP_PError) ) {
		  cout << "PError LOW!" << endl;
		  return;
	    }

	    if ( ! (pins & BPP_GP_nAck) ) {
		  cout << "nAck LOW!" << endl;
		  return;
	    }

	    if ( pins & BPP_GP_Busy ) {
		  cout << "Busy HIGH!" << endl;
		  return;
	    }

	    if ( ! (pins & BPP_GP_Select) ) {
		  cout << "Select LOW!" << endl;
		  return;
	    }

	    cout << "OK" << endl;

	    unsigned long data = strtoul(argv[idx], 0, 0);
	    data &= 0xff;

	    rc = ioctl(device, BPP_SET_INPUT, 0);
	    rc = ioctl(device, BPP_PUT_DATA, data);
	    pins |= BPP_PP_nAutoFd;
	    rc = ioctl(device, BPP_PUT_PINS, pins);
	    cout << "Event 34: DONE" << endl;

	    pins &= ~BPP_PP_nStrobe;
	    rc = ioctl(device, BPP_PUT_PINS, pins);
	    assert( rc != -1 );
	    cout << "Event 35: DONE" << endl;

	    cout << "Event 36: " << flush;
	    pins = ioctl(device, BPP_GET_PINS);

	    if ( ! (pins & BPP_GP_PError) ) {
		  cout << "PError LOW!" << endl;
		  return;
	    }

	    if ( ! (pins & BPP_GP_nAck) ) {
		  cout << "nAck LOW!" << endl;
		  return;
	    }

	    if ( ! (pins & BPP_GP_Busy) ) {
		  cout << "Busy LOW!" << endl;
		  return;
	    }

	    if ( ! (pins & BPP_GP_Select) ) {
		  cout << "Select LOW!" << endl;
		  return;
	    }

	    cout << "OK" << endl;

	    pins |= BPP_PP_nStrobe;
	    rc = ioctl(device, BPP_PUT_PINS, pins);
	    cout << "Event 37: DONE" << endl;

	    cout << "Event 32: " << flush;
	    pins = ioctl(device, BPP_GET_PINS);

	    if ( ! (pins & BPP_GP_PError) ) {
		  cout << "PError LOW!" << endl;
		  return;
	    }

	    if ( ! (pins & BPP_GP_nAck) ) {
		  cout << "nAck LOW!" << endl;
		  return;
	    }

	    if ( pins & BPP_GP_Busy ) {
		  cout << "Busy HIGH!" << endl;
		  return;
	    }

	    if ( ! (pins & BPP_GP_Select) ) {
		  cout << "Select LOW!" << endl;
		  return;
	    }

	    cout << "OK" << endl;

      }
}

static void nibble(unsigned, char*[])
{
      int rc;
      int pins = ioctl(device, BPP_GET_PINS);

      cout << "--> Negotiate nibble mode." << endl;

      cout << "Event 0: ";
      if ( (pins & PP_MASK) != (BPP_PP_nInit|BPP_PP_nStrobe|BPP_PP_nAutoFd) ) {
	    cout << "Not in compatibility mode!" << endl;
	    return;
      }
      cout << "OK" << endl;

        // Perform event 1
      rc = ioctl(device, BPP_PUT_DATA, 0x00);
      assert( rc != -1 );
      pins = (pins | BPP_PP_nSelectIn) & ~BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      assert( rc != -1 );
      cout << "Event 1: DONE" << endl;

        // Watch for Event 2 response from the peripheral
      cout << "Event 2: " << flush;
      pins = ioctl(device, BPP_GET_PINS);
      if ( ! (pins & BPP_GP_PError) ) {
	    cout << "PError LOW!" << endl;
	    return;
      }

      if (pins & BPP_GP_nAck) {
	    cout << "nAck HIGH!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_nFault) ) {
	    cout << "nFault LOW!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_Select) ) {
	    cout << "Select LOW!" << endl;
	    return;
      }

      cout << "OK" << endl;

        // Perform Event 3
      pins &= ~BPP_PP_nStrobe;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      cout << "Event 3: DONE" << endl;

        // Perform Event 4
      pins |= BPP_PP_nStrobe|BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      cout << "Event 4: DONE" << endl;

	// Event 5 is transient.
      cout << "(Event 5 is transient)" << endl;

        // Watch for Event 6
      cout << "Event 6: " << flush;
      pins = ioctl(device, BPP_GET_PINS);

      if (pins & BPP_GP_Select) {
	    cout << "Select HIGH!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_nAck) ) {
	    cout << "nAck LOW!" << endl;
	    return;
      }

      if ( (0 == (pins & BPP_GP_PError)) != (0 ==  (pins & BPP_GP_nFault)) ) {
	    cout << "PError and nFault don't match!" << endl;
	    return;
      }

      if (pins & BPP_GP_nFault)
	cout << "(nFault high, no data) ";
      else
	cout << "(nFault low, data available) ";

      cout << "OK" << endl;
}

static void terminate(unsigned, char*[])
{
      int rc;
      int pins = ioctl(device, BPP_GET_PINS);
      const int select_level = pins & BPP_GP_Select;

      assert( pins != -1 );

      if ( ! (pins & BPP_PP_nSelectIn) ) {
	    cout << "1284 is not active." << endl;
	    return;
      }

      cout << "--> Termination Phase" << endl;

      pins &= ~BPP_PP_nSelectIn;
      pins |= BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      cout << "Event 22: DONE" << endl;

      cout << "(Event 23 is transient)" << endl;

      cout << "Event 24: " << flush;
      pins = ioctl(device, BPP_GET_PINS);
      assert( pins != -1 );

      if (pins & BPP_GP_nAck) {
	    cout << "nAck HIGH!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_Busy) ) {
	    cout << "Busy LOW!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_nFault) ) {
	    cout << "nFault LOW!" << endl;
	    return;
      }

      if (select_level == (pins & BPP_GP_Select)) {
	    cout << "Select INCORRECT!" << endl;
	    return;
      }

      cout << "OK" << endl;

      pins &= ~BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      assert( rc != -1 );
      cout << "Event 25: DONE" << endl;

      cout << "(Event 26 is transient)" << endl;

      cout << "Event 27: " << flush;
      pins = ioctl(device, BPP_GET_PINS);
      assert( pins != -1 );

      if ( ! (pins & BPP_GP_nAck) ) {
	    cout << "nAck LOW!" << endl;
	    return;
      }

      if ( ! (pins & BPP_GP_Busy) ) {
	    cout << "Busy LOW!" << endl;
	    return;
      }

      if (pins & BPP_GP_nFault)
	cout << "(nFault HIGH) ";

      if (pins & BPP_GP_Select)
	cout << "(Select HIGH) ";

      cout << "OK" << endl;

      pins |= BPP_PP_nAutoFd;
      rc = ioctl(device, BPP_PUT_PINS, pins);
      assert( rc != -1 );
      cout << "Event 28: DONE" << endl;

}

static void read_port(unsigned argc, char*argv[])
{
      char buffer[1];

      int rc = read(device, buffer, sizeof buffer);
      if (rc == -1) {
	    cout << "read rc==" << rc << " errno=" << errno << endl;
      } else {
	    cout << " Read rc==" << rc << " -->" << buffer[0] << "<--" << endl;
      }
}

struct {
      const char*name;
      cmd_fun fun;
} table[] = {
      { "change", changepins },
      { "data",  datapins },
      { "init",  initpins },
      { "set",   setpins },
      { "show",  showpins },
      { "quit",  quit },
      { "input", inout },
      { "output",inout },
      { "ECP",   ECP },
      { "ecpsend", ecpsend },
      { "terminate", terminate },
      { "nibble", nibble },
      { "read",  read_port },
      { "help",  help },
      { "?",     help },
      { 0,       no_such_command }
};


cmd_fun lookup(const char*name)
{
      unsigned idx;
      for (idx = 0 ;  table[idx].name ;  idx += 1)
	if (strcmp(name, table[idx].name) == 0) break;

      return table[idx].fun;
}
